# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================


from typing import Any

from pydantic import BaseModel
from typing_extensions import Self

from pixano.datasets import Dataset, DatasetItem, DatasetSchema
from pixano.features import Annotation, Entity, Item, View
from pixano.features.schemas.base_schema import BaseSchema
from pixano.features.schemas.registry import _PIXANO_SCHEMA_REGISTRY
from pixano.features.schemas.schema_group import SchemaGroup
from pixano.utils.python import get_super_type_from_dict

from .annotations import AnnotationModel
from .base_schema import BaseSchemaModel
from .entities import EntityModel
from .items import ItemModel
from .table_info import TableInfo
from .views import ViewModel


class DatasetItemModel(BaseModel):
    """[DatasetItem][pixano.datasets.DatasetItem] model.

    It represents a dataset item with its associated entities, annotations and views.

    The mappings consist of the table name as key and the corresponding model or list of models as value.

    Attributes:
        id: The dataset item id.
        item: The item model.
        entities: The entities models mapping.
        annotations: The annotations models mapping.
        views: The views models mapping.
    """

    id: str
    item: ItemModel
    entities: dict[str, list[EntityModel] | EntityModel | None] = {}
    annotations: dict[str, list[AnnotationModel] | AnnotationModel | None] = {}
    views: dict[str, list[ViewModel] | ViewModel | None] = {}

    def model_dump(self, exclude_timestamps: bool = False, **kwargs: Any) -> dict[str, Any]:
        """Dump the model to a dictionary.

        Args:
            exclude_timestamps: Exclude timestamps "created_at" and "updated_at" from the model dump. Useful for
                comparing models without timestamps.
            kwargs: Arguments for pydantic `BaseModel.model_dump()`.

        Returns:
            The model dump.
        """
        model_dump = super().model_dump(**kwargs)
        if exclude_timestamps:
            model_dump["item"].pop("created_at", None)
            model_dump["item"].pop("updated_at", None)
            for k in ["entities", "annotations", "views"]:
                for model in model_dump[k].values():
                    if model is None:
                        continue
                    elif isinstance(model, list):  # Only one level deep.
                        for item in model:
                            item.pop("created_at", None)
                            item.pop("updated_at", None)
                    else:
                        model.pop("created_at", None)
                        model.pop("updated_at", None)
        return model_dump

    @classmethod
    def from_dataset_item(cls, dataset_item: DatasetItem, dataset_schema: DatasetSchema) -> Self:
        """Create a model from a [DatasetItem][pixano.datasets.DatasetItem].

        Args:
            dataset_item: The dataset item to create the model from.
            dataset_schema: The schema of the dataset containing the dataset item.

        Returns:
            The created model.
        """

        def _row_or_rows_to_model_or_models(
            row_or_rows: BaseSchema | list[BaseSchema], name: str, group: SchemaGroup, model: type[BaseSchemaModel]
        ) -> BaseSchemaModel | list[BaseSchemaModel]:
            base_schema = get_super_type_from_dict(
                type(row_or_rows[0]) if isinstance(row_or_rows, list) else type(row_or_rows), _PIXANO_SCHEMA_REGISTRY
            )
            if base_schema is None:
                raise ValueError(f"Unsupported schema type {type(row_or_rows)}")
            table_info = TableInfo(name=name, group=group.value, base_schema=base_schema.__name__)
            if isinstance(row_or_rows, list):
                return model.from_rows(row_or_rows, table_info=table_info)
            else:
                return model.from_row(row_or_rows, table_info=table_info)

        model_dict: dict[str, Any] = {
            "entities": {},
            "annotations": {},
            "views": {},
        }
        for key, value in dataset_item.to_schemas_data(dataset_schema).items():
            if value is None or value == []:
                if issubclass(dataset_schema.schemas[key], View):
                    model_dict["views"][key] = None if value is None else []
                elif issubclass(dataset_schema.schemas[key], Entity):
                    model_dict["entities"][key] = None if value is None else []
                elif issubclass(dataset_schema.schemas[key], Annotation):
                    model_dict["annotations"][key] = None if value is None else []
                else:
                    raise ValueError(f"Unsupported schema type {type(value)}")
            elif isinstance(value, Item) or isinstance(value, list) and isinstance(value[0], Item):
                model_dict[key] = _row_or_rows_to_model_or_models(value, key, SchemaGroup.ITEM, ItemModel)
            elif isinstance(value, Annotation) or isinstance(value, list) and isinstance(value[0], Annotation):
                model_dict["annotations"][key] = _row_or_rows_to_model_or_models(
                    value, key, SchemaGroup.ANNOTATION, AnnotationModel
                )
            elif isinstance(value, Entity) or isinstance(value, list) and isinstance(value[0], Entity):
                model_dict["entities"][key] = _row_or_rows_to_model_or_models(
                    value, key, SchemaGroup.ENTITY, EntityModel
                )
            elif isinstance(value, View) or isinstance(value, list) and isinstance(value[0], View):
                model_dict["views"][key] = _row_or_rows_to_model_or_models(value, key, SchemaGroup.VIEW, ViewModel)
            else:
                raise ValueError(f"Unsupported schema type {type(value)}")
        model_dict["id"] = dataset_item.id
        return cls.model_validate(model_dict)

    @classmethod
    def from_dataset_items(cls, dataset_items: list[DatasetItem], dataset_schema: DatasetSchema) -> list[Self]:
        """Create a list of models from a list of [DatasetItem][pixano.datasets.DatasetItem]s.

        Args:
            dataset_items: The dataset items to create the models from.
            dataset_schema: The schema of the dataset containing the dataset item.

        Returns:
            The list of created models.
        """
        return [cls.from_dataset_item(dataset_item, dataset_schema) for dataset_item in dataset_items]

    def to_dataset_item(self, dataset: Dataset) -> DatasetItem:
        """Create a [DatasetItem][pixano.datasets.DatasetItem] from a model.

        Args:
            dataset: The dataset containing the model.

        Returns:
            The created dataset item.
        """
        schema_dict = {}

        item = self.item
        schema_dict.update(item.to_row(dataset).model_dump())

        for group in [self.annotations, self.entities, self.views]:
            for key, value in group.items():
                if isinstance(value, list):
                    schema_dict[key] = [v.to_row(dataset) for v in value]
                elif value is None:
                    schema_dict[key] = None
                else:
                    schema_dict[key] = value.to_row(dataset)

        return DatasetItem.from_dataset_schema(dataset.schema, exclude_embeddings=True).model_validate(schema_dict)

    @staticmethod
    def to_dataset_items(models: list["DatasetItemModel"], dataset: Dataset) -> list[DatasetItem]:
        """Create a list of [DatasetItem][pixano.datasets.DatasetItem]s from a list of models.

        Args:
            models: The models to create the dataset items from.
            dataset: The dataset containing the model.

        Returns:
            The list of created dataset items.
        """
        return [model.to_dataset_item(dataset) for model in models]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================


from pixano_inference.pydantic import NDArrayFloat

from pixano.utils import issubclass_strict


def is_ndarray_float(cls: type, strict: bool = False) -> bool:
    """Check if a class is a `NDArrayFloat` or a subclass of `NDArrayFloat`."""
    return issubclass_strict(cls, NDArrayFloat, strict)


def create_ndarray_float(values: list[float], shape: list[int]) -> NDArrayFloat:
    """Create a `NDArrayFloat` instance.

    Args:
        values: The list of floating-point values in the array.
        shape: The shape of the array, represented as a list of integers.

    Returns:
        The created `NDArrayFloat` instance.
    """
    return NDArrayFloat(values=values, shape=shape)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing_extensions import TYPE_CHECKING

from pixano.utils import issubclass_strict

from ...types.schema_reference import EntityRef, ItemRef, ViewRef
from ..base_schema import BaseSchema
from ..registry import _register_schema_internal


if TYPE_CHECKING:
    from pixano.features.schemas import Item, View


@_register_schema_internal
class Entity(BaseSchema):
    """`Entity` base class.

    Entities are used to define an entity in a dataset such as an object, a track. It can refer to an item, a view,
    and a parent entity.

    Attributes:
        item_ref: Reference to the entity's item.
        view_ref: Reference to the entity's view.
        parent_ref: Reference to the entity's parent entity.
    """

    item_ref: ItemRef = ItemRef.none()
    view_ref: ViewRef = ViewRef.none()
    parent_ref: EntityRef = EntityRef.none()

    @property
    def item(self) -> "Item":
        """Get the entity's item."""
        return self.resolve_ref(self.item_ref)

    @property
    def view(self) -> "View":
        """Get the entity's view."""
        return self.resolve_ref(self.view_ref)

    @property
    def parent(self) -> "Entity":
        """Get the entity's parent entity."""
        return self.resolve_ref(self.parent_ref)


def is_entity(cls: type, strict: bool = False) -> bool:
    """Check if a class is an Entity or subclass of Entity."""
    return issubclass_strict(cls, Entity, strict)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import logging


logging.basicConfig(level=logging.INFO)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import html
import json

import IPython.display
import shortuuid


def display_colab(url: str, port: int, height: int):
    """Display a Pixano app inside a Google Colab.

    Args:
        url: Pixano app URL.
        port: Pixano app port.
        height: Frame height.
    """
    # Define frame template
    shell = """
        (() => {
            const url = new URL(%URL%);
            const port = %PORT%;
            if (port) {
                url.port = port;
            }
            const iframe = document.createElement('iframe');
            iframe.src = url;
            iframe.setAttribute('width', '100%');
            iframe.setAttribute('height', '%HEIGHT%');
            iframe.setAttribute('frameborder', 0);
            document.body.appendChild(iframe);
        })();
    """

    # Replace variables in template
    replacements = [
        ("%HEIGHT%", f"{height}"),
        ("%PORT%", f"{port}"),
        ("%URL%", json.dumps(url)),
    ]
    for k, v in replacements:
        shell = shell.replace(k, v)

    # Display frame
    script = IPython.display.Javascript(shell)
    IPython.display.display(script)


def display_ipython(url: str, port: int, height: int):
    """Display a Pixano app inside a Jupyter notebook.

    Args:
        url: Pixano app URL.
        port: Pixano app port.
        height: Frame height.
    """
    # Define frame template
    shell = """
        <iframe id="%HTML_ID%" width="100%" height="%HEIGHT%" frameborder="0">
        </iframe>
        <script>
            (function() {
                const frame = document.getElementById(%JSON_ID%);
                const url = new URL(%URL%, window.location);
                const port = %PORT%;
                if (port) {
                    url.port = port;
                }
                frame.src = url;
            })();
        </script>
    """

    # Replace variables in template
    frame_id = f"frame-{shortuuid.uuid()}"
    replacements = [
        ("%HTML_ID%", html.escape(frame_id, quote=True)),
        ("%JSON_ID%", json.dumps(frame_id)),
        ("%HEIGHT%", f"{height}"),
        ("%PORT%", f"{port}"),
        ("%URL%", json.dumps(url)),
    ]
    for k, v in replacements:
        shell = shell.replace(k, v)

    # Display frame
    iframe = IPython.display.HTML(shell)
    IPython.display.display(iframe)


def display_cli(url: str, port: int):
    """Display a Pixano app inside a command line interface.

    Args:
        url: Pixano app URL.
        port: Pixano app port.
    """
    print(f"Please visit {url}:{port} in a web browser.")



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import shortuuid
from fastapi import HTTPException
from pixano_inference.client import PixanoInferenceClient

from pixano.app.settings import Settings
from pixano.datasets import Dataset
from pixano.features import Source


def get_client_from_settings(settings: Settings) -> PixanoInferenceClient:
    """Get the Pixano inference client from the settings."""
    client = settings.pixano_inference_client
    if client is None:
        raise HTTPException(status_code=500, detail="PixanoInferenceClient not set in settings")
    return client


def get_model_source(dataset: Dataset, model: str):
    """Get the model's source from a given Dataset and Model.

    If it exists in the database already it returns that one otherwise creates a new instance.
    """
    sources: list[Source] = dataset.get_data(table_name="source", limit=2, where=f"name='{model}' AND kind='model'")
    # TODO: enforce check consistency for source to have a unique name instead of checking here.
    if len(sources) > 1:
        raise HTTPException(status_code=400, detail="Only one source for model is allowed.")
    elif len(sources) == 0:
        source = Source(id=shortuuid.uuid(), name=model, kind="model")
        dataset.add_data("source", data=[source])
    else:
        source = sources[0]
    return source



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from __future__ import annotations

import shutil
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, cast, overload

import lancedb
import polars as pl
import pyarrow as pa
from lancedb.common import DATA
from lancedb.table import LanceTable
from pydantic import ConfigDict

from pixano.datasets.queries import TableQueryBuilder
from pixano.datasets.utils.errors import DatasetAccessError, DatasetPaginationError
from pixano.datasets.utils.integrity import IntegrityCheck, check_table_integrity, handle_integrity_errors
from pixano.features import SchemaGroup, Source, ViewEmbedding, is_view_embedding
from pixano.features.schemas.base_schema import BaseSchema
from pixano.utils.python import to_sql_list

from .dataset_features_values import DatasetFeaturesValues
from .dataset_info import DatasetInfo
from .dataset_schema import DatasetItem, DatasetSchema, SchemaRelation
from .dataset_stat import DatasetStatistic


if TYPE_CHECKING:
    from ..features import (
        Annotation,
        AnnotationRef,
        BaseSchema,
        Embedding,
        EmbeddingRef,
        Entity,
        EntityRef,
        Item,
        ItemRef,
        SchemaRef,
        SourceRef,
        View,
        ViewRef,
    )


def _validate_ids_and_limit_and_skip(ids: list[str] | None, limit: int | None, skip: int = 0) -> None:
    if ids is None and limit is None:
        raise DatasetPaginationError("limit must be set if ids is None")
    elif ids is not None and limit is not None:
        raise DatasetPaginationError("ids and limit cannot be set at the same time")
    elif ids is not None and (not isinstance(ids, list) or not all(isinstance(i, str) for i in ids)):
        raise DatasetPaginationError("ids must be a list of strings")
    elif limit is not None and (not isinstance(limit, int) or limit < 0) or not isinstance(skip, int) or skip < 0:
        raise DatasetPaginationError("limit and skip must be positive integers")


def _validate_ids_item_ids_and_limit_and_skip(
    ids: list[str] | None, limit: int | None, skip: int = 0, item_ids: list[str] | None = None
) -> None:
    if ids is not None and item_ids is not None:
        raise DatasetPaginationError("ids and item_ids cannot be set at the same time")
    if ids is not None and limit is not None:
        raise DatasetPaginationError("ids and limit cannot be set at the same time")
    elif ids is not None and (not isinstance(ids, list) or not all(isinstance(i, str) for i in ids)):
        raise DatasetPaginationError("ids must be a list of strings")
    elif item_ids is not None and (not isinstance(item_ids, list) or not all(isinstance(i, str) for i in item_ids)):
        raise DatasetPaginationError("item_ids must be a list of strings")
    elif limit is not None and (not isinstance(limit, int) or limit < 1) or not isinstance(skip, int) or skip < 0:
        raise DatasetPaginationError("limit and skip must be positive integers")


def _validate_raise_or_warn(raise_or_warn: Literal["raise", "warn", "none"]):
    if raise_or_warn not in ["raise", "warn", "none"]:
        raise ValueError("raise_or_warn must be 'raise', 'warn' or 'none'")


class Dataset:
    """The Pixano Dataset.

    It is a collection of tables that can be queried and manipulated with LanceDB.

    The tables are defined by the [DatasetSchema][pixano.datasets.DatasetSchema] which allows the dataset to return
    the data in the form of [LanceModel][lancedb.pydantic.LanceModel] instances.

    Attributes:
        path: Path to the dataset.
        info: Dataset info.
        schema: Dataset schema.
        features_values: Dataset features values.
        stats: Dataset statistics.
        thumbnail: Dataset thumbnail base 64 URL.
        media_dir: Path to the media directory.
    """

    # Allow arbitrary types because of S3 Path
    model_config = ConfigDict(arbitrary_types_allowed=True)

    _DB_PATH: str = "db"
    _PREVIEWS_PATH: str = "previews"
    _INFO_FILE: str = "info.json"
    _SCHEMA_FILE: str = "schema.json"
    _FEATURES_VALUES_FILE: str = "features_values.json"
    _STAT_FILE: str = "stats.json"
    _THUMB_FILE: str = "preview.png"

    path: Path
    info: DatasetInfo
    schema: DatasetSchema
    features_values: DatasetFeaturesValues = DatasetFeaturesValues()
    stats: list[DatasetStatistic] = []
    thumbnail: Path
    media_dir: Path

    def __init__(self, path: Path, media_dir: Path | None = None):
        """Initialize the dataset.

        Args:
            path: Path to the dataset.
            media_dir: Path to the media directory.
        """
        self.path = path

        self._info_file = self.path / self._INFO_FILE
        self._schema_file = self.path / self._SCHEMA_FILE
        self._features_values_file = self.path / self._FEATURES_VALUES_FILE
        self._stat_file = self.path / self._STAT_FILE
        self._thumb_file = self.path / self._THUMB_FILE
        self._db_path = self.path / self._DB_PATH

        self.info = DatasetInfo.from_json(self._info_file)
        self.features_values = DatasetFeaturesValues.from_json(self._features_values_file)
        self.stats = DatasetStatistic.from_json(self._stat_file) if self._stat_file.is_file() else []
        self.media_dir = media_dir or self.path / "media"
        self.thumbnail = self._thumb_file
        self.previews_path = self.path / self._PREVIEWS_PATH

        self._db_connection = self._connect()

        self._reload_schema()

    def _move_dataset(self, new_path: Path) -> None:
        """Move dataset to a new path.

        Args:
            new_path: New dataset path.
        """
        if self.media_dir == self.path / "media":
            self.media_dir = new_path / "media"

        self.path.rename(new_path)
        self.path = new_path

        self._db_path = self.path / self._DB_PATH
        self._info_file = self.path / self._INFO_FILE
        self._schema_file = self.path / self._SCHEMA_FILE
        self._features_values_file = self.path / self._FEATURES_VALUES_FILE
        self._stat_file = self.path / self._STAT_FILE
        self._thumb_file = self.path / self._THUMB_FILE
        self._db_connection = self._connect()

    def _copy_dataset(self, new_path: Path) -> None:
        """Copy dataset to a new path.

        Args:
            new_path: New dataset path.
        """
        if self.media_dir == self.path / "media":
            self.media_dir = new_path / "media"

        shutil.copytree(self.path, new_path, dirs_exist_ok=True)  # Fine
        self.path = new_path

        self._db_path = self.path / self._DB_PATH
        self._info_file = self.path / self._INFO_FILE
        self._schema_file = self.path / self._SCHEMA_FILE
        self._features_values_file = self.path / self._FEATURES_VALUES_FILE
        self._stat_file = self.path / self._STAT_FILE
        self._thumb_file = self.path / self._THUMB_FILE
        self._db_connection = self._connect()

    @property
    def id(self) -> str:
        """Return the dataset ID."""
        return self.info.id

    @property
    def num_rows(self) -> int:
        """Return the number of rows in the dataset.

        Returns:
            Number of rows.
        """
        # Return number of rows of item table
        return self.open_table(SchemaGroup.ITEM.value).count_rows()

    def _reload_schema(self) -> None:
        """Reload schema.

        Returns:
            DatasetSchema: Dataset schema.
        """
        self.schema: DatasetSchema = DatasetSchema.from_json(self._schema_file)
        self.dataset_item_model: type[DatasetItem] = DatasetItem.from_dataset_schema(
            self.schema, exclude_embeddings=True
        )

    def _connect(self) -> lancedb.db.DBConnection:
        """Connect to dataset with LanceDB.

        Returns:
            Dataset LanceDB connection.
        """
        return lancedb.connect(self._db_path)

    def create_table(
        self,
        name: str,
        schema: type[BaseSchema],
        relation_item: SchemaRelation,
        data: DATA | None = None,
        mode: str = "create",
        exist_ok: bool = False,
        on_bad_vectors: str = "error",
        fill_value: float = 0.0,
    ) -> LanceTable:
        """Add a table to the dataset.

        Args:
            name: Table name.
            schema: Table schema.
            relation_item: Relation with the `'item'` table (table to item).
            data: Table data.
            mode: Table mode ('create', 'overwrite').
            exist_ok: If True, do not raise an error if the table already exists.
            on_bad_vectors: Raise an error, drop or fill bad vectors ("error", "drop", "fill").
            fill_value: Value to fill bad vectors.

        Returns:
            The table created.
        """
        table = self._db_connection.create_table(
            name=name,
            schema=schema,
            data=data,
            mode=mode,
            exist_ok=exist_ok,
            on_bad_vectors=on_bad_vectors,
            fill_value=fill_value,
            embedding_functions=None,
        )
        self.schema.add_schema(name, schema, relation_item, exist_ok or mode == "overwrite")
        self.schema.to_json(self._schema_file)
        self._reload_schema()
        return table

    def open_tables(self, names: list[str] | None = None, exclude_embeddings: bool = True) -> dict[str, LanceTable]:
        """Open the dataset tables with LanceDB.

        Args:
            names: Table names to open. If None, open all tables.
            exclude_embeddings: Whether to exclude embedding tables from the list.

        Returns:
            Dataset tables.
        """
        tables: dict[str, LanceTable] = defaultdict(dict)

        for name in names if names is not None else self.schema.schemas.keys():
            if exclude_embeddings and name in self.schema.groups[SchemaGroup.EMBEDDING]:
                continue
            tables[name] = self.open_table(name)

        return tables

    def open_table(self, name: str) -> LanceTable:
        """Open a dataset table with LanceDB.

        Args:
            name: Name of the table to open.

        Returns:
            Dataset table.
        """
        if name not in self.schema.schemas.keys() and name != SchemaGroup.SOURCE.value:
            raise DatasetAccessError(f"Table {name} not found in dataset")

        table = self._db_connection.open_table(name)
        if name == SchemaGroup.SOURCE.value:
            return table

        schema_table = self.schema.schemas[name]
        if is_view_embedding(schema_table):
            schema_table = cast(type[ViewEmbedding], schema_table)
            try:
                schema_table.get_embedding_fn_from_table(self, name, table.schema.metadata)
            except TypeError:  # no embedding function
                pass
        return table

    @overload
    def resolve_ref(self, ref: ItemRef) -> Item: ...
    @overload
    def resolve_ref(self, ref: ViewRef) -> View: ...
    @overload
    def resolve_ref(self, ref: EmbeddingRef) -> Embedding: ...
    @overload
    def resolve_ref(self, ref: EntityRef) -> Entity: ...
    @overload
    def resolve_ref(self, ref: AnnotationRef) -> Annotation: ...
    @overload
    def resolve_ref(self, ref: SourceRef) -> Source: ...
    @overload
    def resolve_ref(self, ref: SchemaRef) -> BaseSchema: ...
    def resolve_ref(
        self, ref: SchemaRef | ItemRef | ViewRef | EmbeddingRef | EntityRef | AnnotationRef | SourceRef
    ) -> BaseSchema | Item | View | Embedding | Entity | Annotation | Source:
        """Resolve a [SchemaRef][pixano.features.SchemaRef].

        It fetches the data from the table referenced.

        Args:
            ref: Reference to resolve.

        Returns:
            The resolved reference.
        """
        if ref.id == "" or ref.name == "":
            raise DatasetAccessError("Reference should have a name and an id.")
        return self.get_data(ref.name, ids=[ref.id])[0]

    @overload
    def get_data(
        self,
        table_name: str,
        ids: list[str] | None = None,
        limit: int | None = None,
        skip: int = 0,
        where: str | None = None,
        item_ids: list[str] | None = None,
    ) -> list[BaseSchema]: ...
    @overload
    def get_data(
        self,
        table_name: str,
        ids: str,
        limit: int | None = None,
        skip: int = 0,
        where: str | None = None,
        item_ids: None = None,
    ) -> BaseSchema | None: ...

    def get_data(
        self,
        table_name: str,
        ids: list[str] | str | None = None,
        limit: int | None = None,
        skip: int = 0,
        where: str | None = None,
        item_ids: list[str] | None = None,
    ) -> list[BaseSchema] | BaseSchema | None:
        """Read data from a table.

        Data can be filtered by ids, item ids, where clause, or limit and skip.

        Args:
            table_name: Table name.
            where: Where clause.
            ids: ids to read.
            limit: Amount of items to read. If not set, will default to table size.
            skip: The number of data to skip.
            item_ids: Item ids to read.

        Returns:
            List of values.
        """
        if table_name == SchemaGroup.ITEM.value:
            if item_ids is not None:
                if ids is None:
                    ids = item_ids
                else:
                    raise DatasetAccessError("ids and item_ids cannot be set at the same time")
                item_ids = None

        return_list = not isinstance(ids, str)
        ids = [ids] if isinstance(ids, str) else ids

        _validate_ids_item_ids_and_limit_and_skip(ids, limit, skip, item_ids)

        if item_ids is not None:
            sql_item_ids = to_sql_list(item_ids)
        table = self.open_table(table_name)

        if ids is None and item_ids is None and limit is None:
            limit = table.count_rows()

        if ids is None:
            if item_ids is None:
                if where is not None:
                    query = TableQueryBuilder(table).where(where).limit(limit).offset(skip)
                else:
                    query = TableQueryBuilder(table).limit(limit).offset(skip)
            else:
                sql_item_ids = to_sql_list(item_ids)
                if where is not None:
                    where += f" AND item_ref.id IN {sql_item_ids}"
                else:
                    where = f"item_ref.id IN {sql_item_ids}"
                query = TableQueryBuilder(table).where(where).limit(limit).offset(skip)
        else:
            sql_ids = to_sql_list(ids)
            if where is not None:
                where += f" AND id IN {sql_ids}"
            else:
                where = f"id IN {sql_ids}"
            query = TableQueryBuilder(table).where(where)

        schema = self.schema.schemas[table_name] if table_name != SchemaGroup.SOURCE.value else Source

        query_models: list[BaseSchema] = query.to_pydantic(schema)
        for model in query_models:
            model.dataset = self  # type: ignore[attr-defined]
            model.table_name = table_name

        return query_models if return_list else (query_models[0] if query_models != [] else None)

    @overload
    def get_dataset_items(
        self, ids: list[str] | None = None, limit: int | None = None, skip: int = 0
    ) -> list[DatasetItem]: ...
    @overload
    def get_dataset_items(self, ids: str, limit: int | None = None, skip: int = 0) -> DatasetItem | None: ...
    def get_dataset_items(
        self,
        ids: list[str] | str | None = None,
        limit: int | None = None,
        skip: int = 0,
    ) -> list[DatasetItem] | DatasetItem | None:
        """Read dataset items.

        Filter dataset items by ids, or limit and skip.

        Args:
            ids: Item ids to read.
            limit: Amount of items to read.
            skip: The number of data to skip..

        Returns:
            List of dataset items.
        """
        return_list = not isinstance(ids, str)
        ids = [ids] if isinstance(ids, str) else ids

        _validate_ids_and_limit_and_skip(ids, limit, skip)

        items = self.get_data(table_name=SchemaGroup.ITEM.value, where=None, ids=ids, limit=limit, skip=skip)
        if items == []:
            return [] if return_list else None
        item_ids: list[str] = [item.id for item in items]
        sql_ids = to_sql_list(item_ids)

        # Load tables
        ds_tables = self.open_tables(exclude_embeddings=True)

        # Load items data from the tables
        data_dict: dict[str, dict[str, BaseSchema | list[BaseSchema]]] = {item.id: item.model_dump() for item in items}
        for table_name, table in ds_tables.items():
            if table_name == SchemaGroup.ITEM.value:
                continue
            is_collection = self.schema.relations[SchemaGroup.ITEM.value][table_name] == SchemaRelation.ONE_TO_MANY
            table_schema = self.schema.schemas[table_name]

            rows = TableQueryBuilder(table).where(f"item_ref.id in {sql_ids}").to_pydantic(table_schema)

            for row in rows:
                row.dataset = self
                row.table_name = table_name
                item_id = row.item_ref.id
                if is_collection:
                    if table_name not in data_dict[item_id]:
                        data_dict[item_id][table_name] = []
                    data_dict[item_id][table_name].append(row)
                else:
                    data_dict[item_id][table_name] = row

        dataset_items = [self.dataset_item_model(**data_dict[item_id]) for item_id in item_ids]  # type: ignore[arg-type]

        return dataset_items if return_list else (dataset_items[0] if dataset_items != [] else None)

    def find_ids_in_table(self, table_name: str, ids: set[str]) -> dict[str, bool]:
        """Search ids in a table.

        Args:
            table_name: Table name.
            ids: Ids to find.

        Returns:
            Dictionary of ids found. Keys are the ids and values are `True` if the id is found, `False` otherwise.
        """
        if len(ids) == 0:
            return {}
        table = self.open_table(table_name)
        ids_found = list(TableQueryBuilder(table).select(["id"]).where(f"id in {to_sql_list(ids)}").to_polars()["id"])
        return {id: id in ids_found for id in ids}

    def get_all_ids(self, table_name: str = SchemaGroup.ITEM.value) -> list[str]:
        """Get all the ids from a table.

        Args:
            table_name: table to look for ids.

        Returns:
            list of the ids.
        """
        return [row["id"] for row in TableQueryBuilder(self.open_table(table_name)).select(["id"]).to_list()]

    def compute_view_embeddings(self, table_name: str, data: list[dict]) -> None:
        """Compute the [view embeddings][pixano.features.ViewEmbedding] via the
            [Embedding Function][lancedb.embeddings.base.EmbeddingFunction] stored in the table metadata.

        Args:
            table_name: Table name containing the view embeddings.
            data: Data to compute. Dictionary representing a view embedding without the vector field.
        """
        table_schema = self.schema.schemas[table_name]
        if not issubclass(table_schema, ViewEmbedding):
            raise DatasetAccessError(f"Table {table_name} is not a view embedding table")
        if not isinstance(data, list) or not all(isinstance(item, dict) for item in data):
            raise DatasetAccessError("Data must be a list of dictionaries")
        # TODO: improve how to handle shape, this works but feels hacky
        for item in data:
            if "shape" not in item:
                item["shape"] = []
        table = self.open_table(table_name)
        data = pa.Table.from_pylist(
            data, schema=table_schema.to_arrow_schema(remove_vector=True, remove_metadata=True)
        )
        table.add(data)
        return None

    def add_data(
        self,
        table_name: str,
        data: list[BaseSchema],
        ignore_integrity_checks: list[IntegrityCheck] | None = None,
        raise_or_warn: Literal["raise", "warn", "none"] = "raise",
    ) -> list[BaseSchema]:
        """Add data to a table.

        Args:
            table_name: Table name.
            data: Data to add.
            ignore_integrity_checks: List of integrity checks to ignore.
            raise_or_warn: Whether to raise or warn on integrity errors. Can be 'raise', 'warn' or 'none'.
        """
        if not all((isinstance(item, type(data[0])) for item in data)) or not set(
            type(data[0]).model_fields.keys()
        ) == set(
            self.schema.schemas[table_name].model_fields.keys()
            if table_name != SchemaGroup.SOURCE.value
            else Source.model_fields.keys()
        ):
            raise DatasetAccessError(
                "All data must be instances of the table type "
                f"{self.schema.schemas[table_name] if table_name != SchemaGroup.SOURCE.value else Source}."
            )
        _validate_raise_or_warn(raise_or_warn)

        table = self.open_table(table_name)
        if raise_or_warn != "none":
            handle_integrity_errors(
                check_table_integrity(table_name, self, data, False, ignore_integrity_checks), raise_or_warn
            )
        for d in data:
            d.created_at = datetime.now()
            d.updated_at = d.created_at
        table.add(data)

        return data

    @overload
    def add_dataset_items(self, dataset_items: DatasetItem) -> DatasetItem: ...
    @overload
    def add_dataset_items(self, dataset_items: list[DatasetItem]) -> list[DatasetItem]: ...
    def add_dataset_items(self, dataset_items: list[DatasetItem] | DatasetItem) -> list[DatasetItem] | DatasetItem:
        """Add dataset items to the dataset.

        Warn:
            Does not test for integrity of the data.

        Args:
            dataset_items: Dataset items to add.
        """
        batch = True
        if isinstance(dataset_items, DatasetItem):
            dataset_items = [dataset_items]
            batch = False
        fields = self.dataset_item_model.model_fields.keys()
        if not all(
            isinstance(item, DatasetItem) and set(fields) == set(item.model_fields.keys()) for item in dataset_items
        ):
            raise DatasetAccessError("All data must be instances of the same DatasetItem.")

        schemas_data = [item.to_schemas_data(self.schema) for item in dataset_items]
        tables_data: dict[str, Any] = {}
        for table_name in self.schema.schemas.keys():
            for item in schemas_data:
                if table_name not in tables_data:
                    tables_data[table_name] = []
                if table_name not in item:
                    continue
                if isinstance(item[table_name], list):
                    tables_data[table_name].extend(item[table_name])
                elif item[table_name] is not None:
                    tables_data[table_name].append(item[table_name])
        for table_name, table_data in tables_data.items():
            if table_data != []:
                self.add_data(
                    table_name=table_name,
                    data=table_data,
                    ignore_integrity_checks=[],
                    raise_or_warn="none",
                )
        return dataset_items if batch else dataset_items[0]

    def delete_data(self, table_name: str, ids: list[str]) -> list[str]:
        """Delete data from a table.

        Args:
            table_name: Table name.
            ids: Ids to delete.

        Returns:
            The list of ids not found.
        """
        if not isinstance(ids, list) or not all(isinstance(i, str) for i in ids):
            raise DatasetAccessError("ids must be a list of strings")

        set_ids = set(ids)

        table = self.open_table(table_name)
        sql_ids = to_sql_list(set_ids)

        ids_found = {
            row["id"] for row in TableQueryBuilder(table).select(["id"]).where(f"id in {to_sql_list(ids)}").to_list()
        }
        ids_not_found = [id for id in set_ids if id not in ids_found]

        table.delete(where=f"id in {sql_ids}")

        return ids_not_found

    def delete_dataset_items(self, ids: list[str]) -> list[str]:
        """Delete dataset items.

        Args:
            ids: Ids to delete.

        Returns:
            The list of ids not found.
        """
        sql_ids = to_sql_list(ids)

        ids_not_found = []
        for table_name in self.schema.schemas.keys():
            if table_name == SchemaGroup.ITEM.value:
                ids_not_found = self.delete_data(table_name, ids)
            else:
                table = self.open_table(table_name)
                table_ids = (
                    table.search()
                    .select(["id"])
                    .where(f"item_ref.id in {sql_ids}")
                    .limit(None)
                    .to_arrow()["id"]
                    .to_pylist()
                )
                if table_ids == []:
                    continue
                table_sql_ids = to_sql_list(table_ids)
                table.delete(where=f"id in {table_sql_ids}")
        return ids_not_found

    @overload
    def update_data(
        self,
        table_name: str,
        data: list[BaseSchema],
        return_separately: Literal[False] = False,
        ignore_integrity_checks: list[IntegrityCheck] | None = None,
        raise_or_warn: Literal["raise", "warn", "none"] = "raise",
    ) -> list[BaseSchema]: ...
    @overload
    def update_data(
        self,
        table_name: str,
        data: list[BaseSchema],
        return_separately: Literal[True],
        ignore_integrity_checks: list[IntegrityCheck] | None = None,
        raise_or_warn: Literal["raise", "warn", "none"] = "raise",
    ) -> tuple[list[BaseSchema], list[BaseSchema]]: ...
    def update_data(
        self,
        table_name: str,
        data: list[BaseSchema],
        return_separately: bool = False,
        ignore_integrity_checks: list[IntegrityCheck] | None = None,
        raise_or_warn: Literal["raise", "warn", "none"] = "raise",
    ) -> list[BaseSchema] | tuple[list[BaseSchema], list[BaseSchema]]:
        """Update data in a table.

        Args:
            table_name: Table name.
            data: Data to update.
            return_separately: Whether to return separately added and updated data.
            ignore_integrity_checks: List of integrity checks to ignore.
            raise_or_warn: Whether to raise or warn on integrity errors. Can be 'raise', 'warn' or 'none'.

        Returns:
            If `return_separately` is `True`, returns a tuple of updated and added data. Otherwise, returns the updated
            data.
        """
        if not all((isinstance(item, type(data[0])) for item in data)) or not set(
            type(data[0]).model_fields.keys()
        ) == set(
            self.schema.schemas[table_name].model_fields.keys()
            if table_name != SchemaGroup.SOURCE.value
            else Source.model_fields.keys()
        ):
            raise DatasetAccessError(
                "All data must be instances of the table type "
                f"{self.schema.schemas[table_name] if table_name != SchemaGroup.SOURCE.value else Source}."
            )
        _validate_raise_or_warn(raise_or_warn)

        table = self.open_table(table_name)
        if raise_or_warn != "none":
            handle_integrity_errors(
                check_table_integrity(table_name, self, data, True, ignore_integrity_checks), raise_or_warn
            )
        set_ids = {item.id for item in data}
        ids_found: dict[str, datetime] = {}

        ids_found = {
            row["id"]: row["created_at"]
            for row in TableQueryBuilder(table)
            .select(["id", "created_at"])
            .where(f"id in {to_sql_list(set_ids)}")
            .to_list()
        }

        for d in data:
            d.updated_at = datetime.now()
            if d.id not in ids_found:
                d.created_at = d.updated_at
        table.merge_insert("id").when_matched_update_all().when_not_matched_insert_all().execute(data)

        if not return_separately:
            return data

        updated_data, added_data = [], []
        for d in data:
            if d.id not in ids_found:
                added_data.append(d)
            else:
                updated_data.append(d)

        return updated_data, added_data

    @overload
    def update_dataset_items(
        self,
        dataset_items: list[DatasetItem],
        return_separately: Literal[False] = False,
    ) -> list[DatasetItem]: ...
    @overload
    def update_dataset_items(
        self,
        dataset_items: list[DatasetItem],
        return_separately: Literal[True],
    ) -> tuple[list[DatasetItem], list[DatasetItem]]: ...
    def update_dataset_items(
        self,
        dataset_items: list[DatasetItem],
        return_separately: bool = False,
    ) -> list[DatasetItem] | tuple[list[DatasetItem], list[DatasetItem]]:
        """Update dataset items.

        Warn:
            Does not test for integrity of the data.

        Args:
            dataset_items: Dataset items to update.
            return_separately: Whether to return separately added and updated dataset items.

        Returns:
            If `return_separately` is `True`, returns a tuple of updated and added dataset items. Otherwise, returns
            the updated dataset items.
        """
        fields = self.dataset_item_model.model_fields.keys()
        if not all(
            isinstance(item, DatasetItem) and set(fields) == set(item.model_fields.keys()) for item in dataset_items
        ):
            raise DatasetAccessError("All data must be instances of the same DatasetItem.")

        schemas_data = [item.to_schemas_data(self.schema) for item in dataset_items]
        updated_ids = set()
        tables_data: dict[str, Any] = {}
        for table_name in self.schema.schemas.keys():
            for item in schemas_data:
                if table_name not in tables_data:
                    tables_data[table_name] = []
                if table_name not in item:
                    continue
                if isinstance(item[table_name], list):
                    tables_data[table_name].extend(item[table_name])
                elif item[table_name] is not None:
                    tables_data[table_name].append(item[table_name])
        for table_name, table_data in tables_data.items():
            if table_data != []:
                updated, _ = self.update_data(
                    table_name,
                    table_data,
                    return_separately=True,
                    ignore_integrity_checks=[],
                    raise_or_warn="none",
                )
                for row in updated:
                    updated_ids.add(row.item_ref.id if table_name != SchemaGroup.ITEM.value else row.id)

        dataset_items = self.get_dataset_items([item.id for item in dataset_items])

        if not return_separately:
            return dataset_items

        updated_items, added_items = [], []
        for item in dataset_items:
            if item.id not in updated_ids:
                added_items.append(item)
            else:
                updated_items.append(item)
        return updated_items, added_items

    @staticmethod
    def find(
        id: str,
        directory: Path,
        media_dir: Path | None = None,
    ) -> "Dataset":
        """Find a Dataset in a directory.

        Args:
            id: Dataset ID to find.
            directory: Directory to search in.
            media_dir: Media directory.

        Returns:
            The found dataset.
        """
        # Browse directory
        for json_fp in directory.glob("*/info.json"):
            info = DatasetInfo.from_json(json_fp)
            if info.id == id:
                # Return dataset
                return Dataset(json_fp.parent, media_dir)
        raise FileNotFoundError(f"Dataset {id} not found in {directory}")

    def semantic_search(
        self, query: str, table_name: str, limit: int, skip: int = 0
    ) -> tuple[list[BaseSchema], list[float]]:
        """Perform a semantic search.

        It searches for the closest items to the query in the table embeddings.

        Args:
            query: Text query for semantic search.
            table_name: Table name for embeddings.
            limit: Limit number of items.
            skip: Skip number of items

        Returns:
            Tuple of items and distances.
        """
        if not isinstance(query, str):
            raise DatasetAccessError("query must be a string.")
        elif not isinstance(table_name, str):
            raise DatasetAccessError("table_name must be a string.")
        elif not isinstance(limit, int) or limit < 1:
            raise DatasetAccessError("limit must be a strictly positive integer.")
        elif not isinstance(skip, int) or skip < 0:
            raise DatasetAccessError("skip must be a positive integer.")
        elif table_name not in self.schema.schemas:
            raise DatasetAccessError(f"Table {table_name} not found in dataset {self.id}.")
        elif table_name not in self.schema.groups[SchemaGroup.EMBEDDING] or not is_view_embedding(
            self.schema.schemas[table_name]
        ):
            raise DatasetAccessError(f"Table {table_name} is not a view embedding table.")

        table = self.open_table(table_name)
        semantic_results: pl.DataFrame = (
            table.search(query).select(["item_ref.id"]).limit(1e9).to_polars()
        )  # TODO: change high limit if lancedb supports it
        item_results = semantic_results.group_by("item_ref.id").agg(pl.min("_distance")).sort("_distance")
        item_ids = item_results["item_ref.id"].to_list()[skip : skip + limit]

        item_rows = self.get_data("item", ids=item_ids)
        item_rows = sorted(item_rows, key=lambda x: item_ids.index(x.id))
        distances = [
            item_results.row(by_predicate=(pl.col("item_ref.id") == item.id), named=True)["_distance"]
            for item in item_rows
        ]
        return item_rows, distances

    @staticmethod
    def list(directory: Path) -> list[DatasetInfo]:
        """List the datasets information in directory.

        Args:
            directory: Directory to search in.

        Returns:
            List of dataset infos.
        """
        dataset_infos = []
        for json_fp in directory.glob("*/info.json"):
            dataset_infos.append(DatasetInfo.from_json(json_fp))
        return dataset_infos



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pathlib import Path
from typing import Any

import shortuuid
from fastapi.encoders import jsonable_encoder
from pixano_inference.client import PixanoInferenceClient
from pixano_inference.pydantic import ImageZeroShotDetectionRequest, ImageZeroShotDetectionResponse
from pixano_inference.utils import is_url

from pixano.features import BBox, Classification, Image, Source
from pixano.features.schemas.entities.entity import Entity
from pixano.features.types.schema_reference import EntityRef, SourceRef, ViewRef


async def image_zero_shot_detection(
    client: PixanoInferenceClient,
    media_dir: Path,
    image: Image,
    entity: Entity,
    source: Source,
    classes: list[str] | str,
    box_threshold: float = 0.5,
    text_threshold: float = 0.5,
    **client_kwargs: Any,
) -> list[tuple[BBox, Classification]]:
    """Image zero shot task.

    Args:
        client: Pixano inference client.
        media_dir: Media directory.
        image: Image to generate mask for.
        entity: Entity associated with the image.
        source: The source refering to the model.
        classes: List of classes to detect in the image.
        box_threshold: Box threshold for detection in the image.
        text_threshold: Text threshold for detection in the image.
        client_kwargs: Additional kwargs for the client to be passed.

    Returns:
        List of BBoxes and Classifications detected in the image with respect to classes and threshold values.
    """
    image_request = image.url if is_url(image.url) else image.open(media_dir, "base64")

    request = ImageZeroShotDetectionRequest(
        image=image_request,
        classes=classes,
        box_threshold=box_threshold,
        text_threshold=text_threshold,
        model=source.name,
    )

    response: ImageZeroShotDetectionResponse = await client.image_zero_shot_detection(request, **client_kwargs)

    inference_metadata = jsonable_encoder(
        {
            "timestamp": response.timestamp,
            "processing_time": response.processing_time,
            **response.metadata,
        }
    )

    boxes = response.data.boxes
    scores = response.data.scores
    detected_classes = response.data.classes

    output: list[tuple[BBox, Classification]] = []

    for b, s, c in zip(boxes, scores, detected_classes, strict=True):
        view_ref = ViewRef(name=image.table_name, id=image.id)
        entity_ref = EntityRef(name=entity.table_name, id=entity.id)
        source_ref = SourceRef(id=source.id)
        output.append(
            (
                BBox(
                    id=shortuuid.uuid(),
                    item_ref=image.item_ref,
                    view_ref=view_ref,
                    entity_ref=entity_ref,
                    source_ref=source_ref,
                    inference_metadata=inference_metadata,
                    coords=b,
                    format="xyxy",
                    is_normalized=False,
                    confidence=s,
                ),
                Classification(
                    id=shortuuid.uuid(),
                    item_ref=image.item_ref,
                    view_ref=view_ref,
                    entity_ref=entity_ref,
                    source_ref=source_ref,
                    inference_metadata=inference_metadata,
                    labels=[c],
                    confidences=[s],
                ),
            )
        )

    return output



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pydantic import model_validator

from pixano.utils import issubclass_strict

from ...types.schema_reference import EntityRef, ItemRef, SourceRef, ViewRef
from ..registry import _register_schema_internal
from .annotation import Annotation


@_register_schema_internal
class Classification(Annotation):
    """Classification at the media level (Image or Text).

    Attributes:
        labels: List of class names.
        confidences: List of prediction confidences.
    """

    labels: list[str]
    confidences: list[float]

    @model_validator(mode="after")
    def _validate_fields(self):
        if len(self.labels) != len(self.confidences):
            raise ValueError("Labels and confidences lists should have the same length")
        return self

    @classmethod
    def none(cls) -> "Classification":
        """Utility function to get a `None` equivalent.
        Should be removed as soon as Lance manages `None` value.

        Returns:
            "None" Classification.
        """
        return cls(
            id="",
            item=ItemRef.none(),
            view=ViewRef.none(),
            entity=EntityRef.none(),
            labels=[],
            confidences=[],
        )

    @property
    def predictions(self) -> list[tuple[str, float]]:
        """Get list of zipped predictions (labels and confidences)."""
        return list(zip(self.labels, self.confidences))


def is_classification(cls: type, strict: bool = False) -> bool:
    """Check if a class is a `Classification` or subclass of `Classification`."""
    return issubclass_strict(cls, Classification, strict)


def create_classification(
    labels: list[str],
    confidences: list[float],
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    view_ref: ViewRef = ViewRef.none(),
    entity_ref: EntityRef = EntityRef.none(),
    source_ref: SourceRef = SourceRef.none(),
) -> Classification:
    """Create a `Classification` instance.

    Args:
        labels: List of class names.
        confidences: List of prediction confidences.
        id: `Classification` ID.
        item_ref: Item reference.
        view_ref: View reference.
        entity_ref: Entity reference.
        source_ref: Source reference.

    Returns:
        The created `Classification` instance.
    """
    return Classification(
        labels=labels,
        confidences=confidences,
        id=id,
        item_ref=item_ref,
        view_ref=view_ref,
        entity_ref=entity_ref,
        source_ref=source_ref,
    )



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from datetime import datetime
from typing import Any

import numpy as np
from PIL import Image as pil_image
from pydantic import field_serializer, field_validator, model_validator
from typing_extensions import Self

from pixano.features.utils import image as image_utils
from pixano.utils import issubclass_strict

from ...types.schema_reference import EntityRef, ItemRef, SourceRef, ViewRef
from ..registry import _register_schema_internal
from .annotation import Annotation


@_register_schema_internal
class CompressedRLE(Annotation):
    """Compressed RLE mask type.

    Attributes:
        size: Mask size.
        counts: Mask RLE encoding.
    """

    size: list[int]
    counts: bytes

    @field_validator("counts", mode="before")
    @classmethod
    def _validate_counts(cls, value: bytes | str) -> bytes:
        if isinstance(value, str):
            value = bytes(value, "utf-8")
        return value

    @model_validator(mode="after")
    def _validate_fields(self):
        if (
            len(self.size) != 2
            and not all(isinstance(s, int) and s > 0 for s in self.size)
            and not (self.size == [0, 0] and self.counts == b"")
        ):
            raise ValueError("Mask size must have 2 elements and be positive integers or [0, 0] for empty mask.")
        return self

    @field_serializer("counts")
    def _serialize_counts(self, value: bytes) -> str:
        return str(value, "utf-8")

    @classmethod
    def none(cls) -> Self:
        """Utility function to get a `None` equivalent.
        Should be removed as soon as Lance manages `None` value.

        Returns:
            "None" `CompressedRLE`.
        """
        return cls(
            id="",
            item=ItemRef.none(),
            view=ViewRef.none(),
            entity=EntityRef.none(),
            size=[0, 0],
            counts=b"",
            created_at=datetime(1970, 1, 1),
            updated_at=datetime(1970, 1, 1),
        )

    @property
    def area(self) -> float:
        """Return mask area.

        Returns:
            Mask area
        """
        return image_utils.mask_area({"size": self.size, "counts": self.counts})

    def to_mask(self) -> np.ndarray:
        """Convert the compressed RLE mask to a NumPy array.

        Returns:
            The mask as a NumPy array.
        """
        return image_utils.rle_to_mask({"size": self.size, "counts": self.counts})

    def to_urle(self) -> dict[str, list[int]]:
        """Convert compressed RLE mask to uncompressed RLE.

        Returns:
            The mask as an uncompressed RLE.
        """
        return image_utils.rle_to_urle(self.model_dump())

    def to_polygons(self) -> list[list]:
        """Convert the compressed RLE mask to poylgons.

        Returns:
            The mask as polygons.
        """
        return image_utils.rle_to_polygons(self.model_dump())

    @staticmethod
    def from_mask(mask: pil_image.Image | np.ndarray, **kwargs: Any) -> "CompressedRLE":
        """Create a compressed RLE mask from a NumPy array.

        Args:
            mask: The mask as a NumPy array.
            kwargs: Additional arguments.

        Returns:
            The compressed RLE mask.
        """
        rle = image_utils.mask_to_rle(mask)
        return CompressedRLE(size=rle["size"], counts=rle["counts"], **kwargs)

    @staticmethod
    def from_urle(urle: dict[str, list[int]], **kwargs: Any) -> "CompressedRLE":
        """Create a compressed RLE mask from an uncompressed RLE.

        Args:
            urle: The mask as an uncompressed RLE.
            kwargs: Additional arguments.

        Returns:
            The compressed RLE mask.
        """
        return CompressedRLE(**image_utils.urle_to_rle(urle), **kwargs)  #  type: ignore[arg-type]

    @staticmethod
    def from_polygons(
        polygons: list[list],
        height: int,
        width: int,
        **kwargs: Any,
    ) -> "CompressedRLE":
        """Create a compressed RLE mask from polygons.

        Args:
            polygons: The mask as polygons.
            height: Image height.
            width: Image width.
            kwargs: Additional arguments.

        Returns:
            The compressed RLE mask.
        """
        return CompressedRLE(**image_utils.polygons_to_rle(polygons, height, width), **kwargs)

    @staticmethod
    def encode(mask: list[list] | dict[str, list[int]], height: int, width: int, **kwargs: Any) -> "CompressedRLE":
        """Create a compressed RLE mask from polygons / uncompressed RLE / compressed RLE.

        Args:
            mask: Mask as polygons / uncompressed RLE / compressed RLE.
            height: Image height.
            width: Image width.
            kwargs: Additional arguments.

        Returns:
            The compressed RLE mask.
        """
        return CompressedRLE(**image_utils.encode_rle(mask, height, width), **kwargs)


def is_compressed_rle(cls: type, strict: bool = False) -> bool:
    """Check if the given class is a subclass of `CompressedRLE`."""
    return issubclass_strict(cls, CompressedRLE, strict)


def create_compressed_rle(
    size: list[int],
    counts: bytes,
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    view_ref: ViewRef = ViewRef.none(),
    entity_ref: EntityRef = EntityRef.none(),
    source_ref: SourceRef = SourceRef.none(),
) -> CompressedRLE:
    """Create a `CompressedRLE` instance.

    Args:
        size: Mask size.
        counts: Mask RLE encoding.
        id: `CompressedRLE` ID.
        item_ref: Item reference.
        view_ref: View reference.
        entity_ref: Entity reference.
        source_ref: Source reference.

    Returns:
        The compressed RLE instance.
    """
    return CompressedRLE(
        size=size,
        counts=counts,
        id=id,
        item_ref=item_ref,
        view_ref=view_ref,
        entity_ref=entity_ref,
        source_ref=source_ref,
    )



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import json
from enum import Enum
from typing import Any, Literal

from pydantic import field_validator, model_validator

from pixano.utils import issubclass_strict

from .base_schema import BaseSchema
from .registry import _register_schema_internal


class SourceKind(Enum):
    """Kind of source that produced the annotation.

    Attributes:
        MODEL: Source produced by a model.
        HUMAN: Source produced by a human.
        GROUND_TRUTH: The source is a ground truth.
        OTHER: Source produced by other means.
    """

    MODEL = "model"
    HUMAN = "human"
    GROUND_TRUTH = "ground_truth"
    OTHER = "other"


@_register_schema_internal
class Source(BaseSchema):
    """Source of the annotation.

    Attributes:
        name: Name of the source.
        kind: Kind of source.
        metadata: Metadata of the source. dict[str, Any] encoded in a string.
    """

    name: str
    kind: str
    metadata: str = json.dumps({})

    @field_validator("kind", mode="before")
    @classmethod
    def _validate_kind_before(cls, v):
        if isinstance(v, SourceKind):
            v = v.value
        return v

    @field_validator("kind")
    @classmethod
    def _validate_kind_after(cls, v):
        kind_values = [kind.value for kind in SourceKind]
        if v not in kind_values:
            raise ValueError(f"Source kind {v} is not valid. Must be one of {kind_values}.")
        return v

    @field_validator("metadata", mode="before")
    @classmethod
    def _validate_metadata(cls, v: str | dict[str, Any]):
        if not isinstance(v, str):
            if not isinstance(v, dict):
                raise ValueError("Metadata must be a dict or string.")
            if not all(isinstance(k, str) for k in v.keys()):
                raise ValueError("Metadata keys must be strings.")
            return json.dumps(v)
        return v

    @field_validator("metadata", mode="after")
    @classmethod
    def _validate_metadata_after(cls, v: str) -> str:
        try:
            json.loads(v)
        except Exception as e:
            raise ValueError("Metadata must be a valid JSON string. Error: " + str(e))
        return v

    @model_validator(mode="after")
    def _validate_fields(self):
        if self.kind == SourceKind.GROUND_TRUTH.value:
            if self.name != "Ground Truth":
                raise ValueError("Ground truth source must have name 'Ground Truth'.")
            if self.id != SourceKind.GROUND_TRUTH.value:
                raise ValueError(f"Ground truth source must have id '{SourceKind.GROUND_TRUTH.value}'.")
        return self

    @classmethod
    def create_ground_truth(cls, metadata: str | dict[str, Any] = {}) -> "Source":
        """Create a ground truth source."""
        return cls(
            id=SourceKind.GROUND_TRUTH.value,
            name="Ground Truth",
            kind=SourceKind.GROUND_TRUTH.value,
            metadata=metadata,
        )


def is_source(cls: type, strict: bool = False) -> bool:
    """Check if a class is a Source or subclass of Source."""
    return issubclass_strict(cls, Source, strict)


def create_source(
    id: str,
    name: str,
    kind: Literal["model", "human", "ground_truth", "other"] | SourceKind,
    metadata: str | dict[str, Any],
) -> Source:
    """Create a `Source` instance.

    Args:
        id: Identifier of the source.
        name: Name of the source.
        kind: Kind of source.
        metadata: Metadata of the source.

    Returns:
        The created `Source` instance.
    """
    return Source(id=id, name=name, kind=kind, metadata=metadata)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from enum import Enum

from pixano.features.schemas.source import Source

from .annotations import Annotation
from .base_schema import BaseSchema
from .embeddings import Embedding
from .entities import Entity
from .items import Item
from .views import View


class SchemaGroup(Enum):
    """Schema group.

    It defines the different schema groups to which a schema can belong.

    Attributes:
        ANNOTATION: Annotation schema group.
        EMBEDDING: Embedding schema group.
        ITEM: Item schema group.
        ENTITY: Entity schema group.
        VIEW: View schema group.
    """

    ANNOTATION = "annotations"
    EMBEDDING = "embeddings"
    ITEM = "item"
    ENTITY = "entities"
    VIEW = "views"
    SOURCE = "source"

    @classmethod
    def _missing_(cls, name):
        for member in cls:
            if member.value == name.lower():
                return member


_SCHEMA_GROUP_TO_SCHEMA_DICT = {
    SchemaGroup.EMBEDDING: Embedding,
    SchemaGroup.ITEM: Item,
    SchemaGroup.ENTITY: Entity,
    SchemaGroup.ANNOTATION: Annotation,
    SchemaGroup.VIEW: View,
    SchemaGroup.SOURCE: Source,
}


def schema_to_group(schema_type: BaseSchema | type) -> SchemaGroup:
    """Get the schema group of a given schema type."""
    try:
        issubclass(schema_type, BaseSchema)
    except TypeError:
        is_class = False
    else:
        is_class = True
    if isinstance(schema_type, Embedding) or is_class and issubclass(schema_type, Embedding):
        return SchemaGroup.EMBEDDING
    elif isinstance(schema_type, Item) or is_class and issubclass(schema_type, Item):
        return SchemaGroup.ITEM
    elif isinstance(schema_type, Entity) or is_class and issubclass(schema_type, Entity):
        return SchemaGroup.ENTITY
    elif isinstance(schema_type, Annotation) or is_class and issubclass(schema_type, Annotation):
        return SchemaGroup.ANNOTATION
    elif isinstance(schema_type, View) or is_class and issubclass(schema_type, View):
        return SchemaGroup.VIEW
    elif isinstance(schema_type, Source) or is_class and issubclass(schema_type, Source):
        return SchemaGroup.SOURCE
    else:
        raise ValueError(f"Unknown schema type: {schema_type}")


def group_to_str(group: SchemaGroup, plural: bool = False) -> str:
    """Convert the schema group to a string.

    Attributes:
        group: The schema group.
        plural: Whether to use the plural form of the word.

    Returns:
        The string for the given schema group.
    """
    if group == SchemaGroup.SOURCE:
        return "sources" if plural else "source"
    elif group == SchemaGroup.ITEM:
        return "items" if plural else "item"
    elif group == SchemaGroup.ENTITY:
        return "entities" if plural else "entity"
    elif group == SchemaGroup.VIEW:
        return "views" if plural else "view"
    elif group == SchemaGroup.ANNOTATION:
        return "annotations" if plural else "annotation"
    elif group == SchemaGroup.EMBEDDING:
        return "embeddings" if plural else "embedding"
    raise ValueError(f"Unknown schema group: {group}")



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import json
from pathlib import Path

from pydantic import BaseModel


class DatasetStatistic(BaseModel):
    """A statistic of a dataset.

    Attributes:
        name: The name of the statistic.
        type: The type ('numerical' or 'categorical') of the statistic.
        histogram: The histogram of the statistic.
        range: The range of the statistic.
    """

    name: str
    type: str
    histogram: list[dict[str, float | int | str]]
    range: list[int | float] | None = None

    @staticmethod
    def from_json(json_fp: Path) -> list["DatasetStatistic"]:
        """Read list of `DatasetStatistic` from a JSON file.

        Args:
            json_fp: JSON file path.

        Returns:
            A list of `DatasetStat`.
        """
        stats_json = json.loads(json_fp.read_text(encoding="utf-8"))

        return [DatasetStatistic.model_validate(stat) for stat in stats_json]

    def to_json(self, json_fp: Path) -> None:
        """Save `DatasetStatistic` to a json file.

        Replace the existing histogram in `json_fp`.

        Args:
            json_fp: Save directory.
        """
        try:
            stats_json = json.loads(json_fp.read_text(encoding="utf-8"))
        except FileNotFoundError:
            stats_json = []
        # keep all stats except the one with same name, we replace it if exist
        stats_json = [stat for stat in stats_json if stat["name"] != self.name]
        stats_json.append({"name": self.name, "type": self.type, "histogram": self.histogram, "range": self.range})

        json_fp.write_text(json.dumps(stats_json, indent=4), encoding="utf-8")



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pathlib import Path

import shortuuid

from pixano.utils import issubclass_strict

from ...types.schema_reference import ItemRef, ViewRef
from ..registry import _register_schema_internal
from .view import View


@_register_schema_internal
class Video(View):
    """Video view.

    Attributes:
        url: The video URL.
        num_frames: The number of frames in the video.
        fps: The frames per second of the video.
        width: The video width.
        height: The video height.
        format: The video format.
        duration: The video duration.
    """

    url: str
    num_frames: int
    fps: float
    width: int
    height: int
    format: str
    duration: float


def is_video(cls: type, strict: bool = False) -> bool:
    """Check if the given class is a `Video` or a subclass of `Video`."""
    return issubclass_strict(cls, Video, strict)


def create_video(
    url: Path,
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    parent_ref: ViewRef = ViewRef.none(),
    num_frames: int | None = None,
    fps: float | None = None,
    width: int | None = None,
    height: int | None = None,
    format: str | None = None,
    duration: float | None = None,
    url_relative_path: Path | None = None,
) -> Video:
    """Create a `Video` instance.

    Args:
        url: The image URL. If not relative, the URL is converted to a relative path using `other_path`.
        id: Point cloud ID.
        item_ref: Item reference.
        parent_ref: Parent view reference.
        num_frames: The number of frames in the video. If None, the number of frames is
            extracted from the video file.
        fps: The frames per second of the video. If None, the fps is extracted from the video
            file.
        width: The video width. If None, the width is extracted from the video file.
        height: The video height. If None, the height is extracted from the video file.
        format: The video format. If None, the format is extracted from the video file.
        duration: The video duration. If None, the duration is extracted from the video file.
        url_relative_path: The path to convert the URL to a relative path.

    Returns:
        The created `Video` instance.
    """
    none_conditions = [
        num_frames is None,
        fps is None,
        width is None,
        height is None,
        format is None,
        duration is None,
    ]
    not_none_conditions = [
        num_frames is not None,
        fps is not None,
        width is not None,
        height is not None,
        format is not None,
        duration is not None,
    ]
    if not all(none_conditions) and not all(not_none_conditions):
        raise ValueError(
            "All or none of the following arguments must be provided: width, height, format, num_frames, fps, duration"
        )

    url = Path(url)
    if id is None:
        id = shortuuid.uuid()

    if width is None:
        try:
            import ffmpeg
        except ImportError:
            raise ImportError("To load video files metadata, install ffmpeg")
        try:
            metadata = ffmpeg.probe(str(url.resolve()), cmd="ffprobe")["streams"][0]
        except FileNotFoundError:
            raise FileNotFoundError("File not found or ffprobe is not installed.")
        r_frame_rate = metadata["r_frame_rate"].split("/")
        fps = float(r_frame_rate[0]) / float(r_frame_rate[1])
        num_frames = int(metadata["nb_frames"])
        width = int(metadata["width"])
        height = int(metadata["height"])
        format = url.suffix[1:]
        duration = float(metadata["duration"])

    if url_relative_path is not None:
        url_relative_path = Path(url_relative_path)
        url = url.relative_to(url_relative_path)

    return Video(
        id=id,
        item_ref=item_ref,
        parent_ref=parent_ref,
        url=url.as_posix(),
        num_frames=num_frames,
        fps=fps,
        width=width,
        height=height,
        format=format,
        duration=duration,
    )



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing import Annotated

from fastapi import APIRouter, Body, Depends, HTTPException

from pixano.app.models import AnnotationModel
from pixano.app.models.entities import EntityModel
from pixano.app.routers.inference.utils import get_client_from_settings
from pixano.app.routers.utils import get_dataset
from pixano.app.settings import Settings, get_settings
from pixano.features import Conversation, Message
from pixano.features.schemas.annotations.text_generation import is_message
from pixano.features.schemas.entities.conversation import is_conversation
from pixano.inference.text_image_conditional_generation import (
    DEFAULT_MAX_NEW_TOKENS,
    DEFAULT_ROLE_ASSISTANT,
    DEFAULT_ROLE_SYSTEM,
    DEFAULT_ROLE_USER,
    DEFAULT_TEMPERATURE,
    text_image_conditional_generation,
)

from .utils import get_model_source


router = APIRouter(prefix="/tasks/conditional_generation", tags=["Task", "Conditional Generation"])


@router.post(
    "/text-image",
    response_model=AnnotationModel,
)
async def call_text_image_conditional_generation(
    dataset_id: Annotated[str, Body(embed=True)],
    conversation: Annotated[EntityModel, Body(embed=True)],
    messages: Annotated[list[AnnotationModel], Body(embed=True)],
    model: Annotated[str, Body(embed=True)],
    settings: Annotated[Settings, Depends(get_settings)],
    max_new_tokens: Annotated[int, Body(embed=True)] = DEFAULT_MAX_NEW_TOKENS,
    temperature: Annotated[float, Body(embed=True)] = DEFAULT_TEMPERATURE,
    role_system: Annotated[str, Body(embed=True)] = DEFAULT_ROLE_SYSTEM,
    role_user: Annotated[str, Body(embed=True)] = DEFAULT_ROLE_USER,
    role_assistant: Annotated[str, Body(embed=True)] = DEFAULT_ROLE_ASSISTANT,
) -> AnnotationModel:
    """Call a text image conditional generation model for a conversation.

    Args:
        dataset_id: The ID of the dataset to use.
        conversation: The conversation to use.
        messages: The messages to use.
        model: The name of the model to use.
        settings: App settings.
        max_new_tokens: The maximum number of tokens to generate.
        temperature: The temperature to use.
        role_system: The role of the system.
        role_user: The role of the user.
        role_assistant: The role of the assistant.

    Returns: The generated message model.
    """
    dataset = get_dataset(dataset_id=dataset_id, dir=settings.library_dir, media_dir=settings.media_dir)
    client = get_client_from_settings(settings=settings)

    if not is_conversation(dataset.schema.schemas[conversation.table_info.name]):
        raise HTTPException(status_code=400, detail="Conversation must be a conversation.")

    conversation_row: Conversation = conversation.to_row(dataset)

    messages_in_one_table = len({m.table_info.name for m in messages}) == 1
    if not messages_in_one_table:
        raise HTTPException(status_code=400, detail="Only one table for messages is allowed.")
    elif not is_message(dataset.schema.schemas[messages[0].table_info.name]):
        raise HTTPException(status_code=400, detail="Messages must be a message.")

    messages_rows: list[Message] = []
    for m in messages:
        m_row: Message = m.to_row(dataset)
        messages_rows.append(m_row)

    source = get_model_source(dataset=dataset, model=model)

    try:
        infered_message = await text_image_conditional_generation(
            client=client,
            source=source,
            dataset=dataset,
            media_dir=settings.media_dir,
            messages=messages_rows,
            conversation=conversation_row,
            max_new_tokens=max_new_tokens,
            temperature=temperature,
            role_system=role_system,
            role_user=role_user,
            role_assistant=role_assistant,
        )
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e)) from e

    message_model = AnnotationModel.from_row(row=infered_message, table_info=messages[0].table_info)

    return message_model



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================


from pixano.datasets.dataset_schema import DatasetItem
from pixano.datasets.workspaces import WorkspaceType
from pixano.datasets.workspaces.dataset_items import DefaultVideoDatasetItem

from .base import FolderBaseBuilder


# TODO: Add more video extensions supported by ffmpeg
VIDEO_EXTENSIONS = [
    ".mp4",
    ".avi",
    ".mov",
    ".mkv",
    ".webm",
    ".flv",
    ".vob",
]


class VideoFolderBuilder(FolderBaseBuilder):
    """Builder for video datasets stored in a folder."""

    EXTENSIONS = VIDEO_EXTENSIONS
    WORKSPACE_TYPE = WorkspaceType.VIDEO
    DEFAULT_SCHEMA: type[DatasetItem] = DefaultVideoDatasetItem



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pixano.utils import issubclass_strict

from ...types.schema_reference import EntityRef, ItemRef, ViewRef
from ..registry import _register_schema_internal
from .entity import Entity


@_register_schema_internal
class NamedEntity(Entity):
    """A named entity.

    A NamedEntity represents an entity with a simple name attribute.

    Attributes:
        name: The name of the entity.
    """

    name: str


def is_named_entity(cls: type, strict: bool = False) -> bool:
    """Check if the given class is a `NamedEntity` or a subclass of `NamedEntity`."""
    return issubclass_strict(cls, NamedEntity, strict)


def create_named_entity(
    name: str,
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    view_ref: ViewRef = ViewRef.none(),
    parent_ref: EntityRef = EntityRef.none(),
) -> NamedEntity:
    """Create a `NamedEntity` instance.

    Args:
        name: The name of the entity.
        id: NamedEntity ID.
        item_ref: Item reference.
        view_ref: View reference.
        parent_ref: Entity reference.

    Returns:
        The created `NamedEntity` instance.
    """
    return NamedEntity(
        id=id,
        item_ref=item_ref,
        view_ref=view_ref,
        parent_ref=parent_ref,
        name=name,
    )
